'###############################################################
'# PBMine.bas                                                  #
'# Written by Ian Cairns, July 26, 2000.                       #
'# In loving memory of my father                               #
'#    John Cairns, July 27, 1924 - June 27, 2000               #
'#                                                             #
'# ian.cairns@gems6.gov.bc.ca                                  #
'#                                                             #
'# Updated 2003 to:                                            #
'#      allow safely exposing extra tiles via F6               #
'#      use a timer to diplay elapsed playing time in seconds  #
'# Any resemblance to any existing game is purely coincidence. #
'# Use and modify at your own risk.                            #
'###############################################################

#COMPILE EXE
#DIM ALL
#INCLUDE "PBMINE.INC"  ' "WIN32API.INC" "RICHEDIT.INC"
#RESOURCE "PBMINE.PBR"

' Define Menu ID codes:
%IDM_NEW       = 1
%IDM_UNDO      = 2
%IDM_OPTIONS   = 3
%IDM_CHEAT     = 4
%IDM_EXIT      = 5
%IDM_HELP      = 6
%IDM_ABOUT     = 7
%IDM_HDRBUTTON = 8
%IDM_TMRBUTTON = 9
' Define Option Dialog Control ID's
%DlgOptions  = 100
%pb_IDOK     = 101
%eWidth      = 102
%eHeight     = 103
%eBombs      = 104

%ID_RICHEDIT      = 200  ' Define RichEdit Dialog ID.
%TIMER_WAIT_VALUE = 200  ' default wait time to check up on display of elapsed time.
%ONE_SECOND       = 10000000  ' 1 second in portions of 100 nanoseconds

' Define Global
%CheatLevel = 3 ' Allow up to 3 tiles to be exposed
                ' at the beginning of the game to get started.
' Define User Messages
%CENTERWINDOW      = 10
%SETVALUE          = %WM_USER + 1
%EXPOSEBLANK       = %WM_USER + 2
%HITBOMB           = %WM_USER + 3
%UNDO              = %WM_USER + 4
%WINNER            = %WM_USER + 5
%USR_CHEAT         = %WM_USER + 6
%USR_ZERO_TIMER    = %WM_USER + 7
%USR_START_TIMER   = %WM_USER + 8
%USR_PAUSE_TIMER   = %WM_USER + 9
%USR_RESTART_TIMER = %WM_USER + 10
'===========================
' Bitmap representing LED bars to be 'lit'
'  -     5
' | |  1   2
'  -     6
' | |  3   4
'  -     7
' Numeric definitions (As bits expressed in binary)"&b MSB - LSB = 76543210"
%Lzero  = &b10111110
%Lone   = &b00010100
%Ltwo   = &b11101100
%Lthree = &b11110100
%Lfour  = &b01010110
%Lfive  = &b11110010
%Lsix   = &b11011010
%Lseven = &b00110100
%Leight = &b11111110
%Lnine  = &b01110110
%Lblank = 0

UNION QuadFILETIME
  dwDateTime AS FILETIME
  qdTime     AS QUAD
END UNION

TYPE PaintBitInfoTYPE
  nWidth       AS LONG  ' width of individual number in pixels
  nHeight      AS LONG  ' height as above
  barThickness AS LONG  ' Thickness of individual LED bar
END TYPE
'===========================
TYPE buttonTYPE
  bHandle  AS DWORD ' Child Window Handle
  bStatus  AS LONG  ' 10 = bomb, 0 = empty square, 1-8 = no of bombs touching this square.
  bDisplay AS LONG  '  0 = Blank, 1 = flag, 2 = ?  3 = flag (X-ed)           //  - Paint on button
END TYPE            '  4 = tile exposed  5 = bomb ON white 6 = bomb ON red   //  - Paint on board

TYPE TileInfoTYPE
  bStatus   AS LONG  ' as for buttonTYPE
  bDisplay  AS LONG  ' as for buttonTYPE
  bUdisplay AS LONG  ' holds old bDisplay values for UNDO
END TYPE

TYPE GameInfoTYPE
  buttonSize AS LONG   ' Assuming buttonWidth = buttonHeight in pixels
  gameCols   AS LONG
  gameRows   AS LONG
  noBombs    AS LONG
  noTiles    AS LONG
  noExposed  AS LONG
  BombCount  AS LONG
  FlagCount  AS LONG
  gameEnd    AS LONG   ' 0 for not end, 1 for Hit bomb but can UNDO, 2 for Winner
END TYPE

'===================
' TileConvertUNION and TileConvertTYPE
' are used to pack/unpack button Information into a Long
' and attach it to the child window's .cbWndExtra
TYPE TileConvertTYPE
  bStatus   AS BYTE
  bDisplay  AS BYTE
  bUdisplay AS BYTE
  bUnused   AS BYTE
END TYPE

UNION TileConvertUNION
  bInfo   AS TileConvertTYPE
  convert AS LONG
END UNION
'===================

GLOBAL szChildClass     AS ASCIIZ * 9, _
       szMenuClass      AS ASCIIZ * 9, _
       gGameBoard()     AS ButtonTYPE, _
       gCheatInfo()     AS LONG, _
       gGameInfo        AS GameInfoTYPE, _
       gNewGameInfo     AS GameInfoTYPE, _
       Kolors()         AS LONG, _
       ghMemDCpic       AS LONG, _  'Memory DC for Display of numbers,flags,bombs
       ghMemDCnum       AS LONG, _  'Memory DC for LED Numbers
       ghMemDCDisp      AS LONG, _  'Memory DC for Time Display
       ghBitmapPic      AS LONG, _  'Bitmap handle for numbers,flags,bombs
       ghOldBitmapPic   AS LONG, _
       ghBitmapNum      AS LONG, _  'Bitmap handle for LED numbers
       ghOldBitmapNum   AS LONG, _
       ghBitmapDisp     AS LONG, _  'Bitmap handle for Time Display
       ghOldBitmapDisp  AS LONG, _
       gMinimumWidth    AS LONG, _  'miminum width of playing field to show timer and button.
       gmenuHt          AS LONG, _
       gTimerRunning    AS LONG      'Flag for message passing

'for RichEdit stream-in..
GLOBAL gPos AS LONG, _
       gPtr AS LONG, _
       gTxt AS STRING
' ========
' Declares
' ========
DECLARE FUNCTION IsWinner AS LONG

DECLARE SUB HelpBoxMaker(BYVAL hParent AS DWORD, BYVAL xPos AS LONG, BYVAL yPos AS LONG, _
                 BYVAL xWidth AS LONG, BYVAL yHeight AS LONG, BYVAL inTitle AS STRING, _
                 BYVAL inText AS STRING)

DECLARE SUB PaintLED(BYVAL inValue AS LONG, BYVAL xWidth AS LONG, BYVAL yHeight AS LONG)

DECLARE CALLBACK FUNCTION HelpBoxProc

'
'------------------------------------
'====================================
FUNCTION WINMAIN (BYVAL hCurrInstance AS LONG, BYVAL hPrevInstance AS LONG,_
                  lpCmdLine AS ASCIIZ PTR, BYVAL nCmdShow AS LONG) AS LONG

  LOCAL  Msg            AS tagMsg,_
         wClass         AS WndClassEx,_
         hWnd           AS DWORD,_
         szClassName    AS ASCIIZ * 7, _
         szIcon         AS ASCIIZ*5

' Initialize
' ~~~~~~~~~~
  szClassName    = "PBMine"
  szChildClass   = "PBButton"
  szmenuClass    = "PBmenu"
  szIcon         = "bomb"
  DIM gGameBoard(0:1, 0:1)

' Fill main window class structure
' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  wClass.cbSize        = SIZEOF(wClass)
  wClass.style         = %CS_OWNDC
  wClass.lpfnWndProc   = CODEPTR(MainWndProc)
  wClass.cbClsExtra    = 0
  wClass.cbWndExtra    = 0
  wClass.hInstance     = hCurrInstance
  wClass.hIcon         = LoadIcon  (hCurrInstance, szIcon)
  wClass.hCursor       = LoadCursor(%NULL, BYVAL %IDC_ARROW)
  wClass.hbrBackground = GetStockObject(%WHITE_BRUSH)
  wClass.lpszMenuName  = %NULL
  wClass.lpszClassName = VARPTR(szClassName)
  wClass.hIConSm       = LoadIcon  (hCurrInstance, szIcon)

' Register the main window-class
' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  CALL RegisterClassEx (wClass) 'register a window class for the program window

' Fill child window class structure
' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  wClass.style         = %NULL
  wClass.lpfnWndProc   = CODEPTR(ButtonProc)
  wClass.cbWndExtra    = 4  ' To hold button Status, Display info, UNDO Display Info.
  wClass.hIcon         = %NULL
  wClass.hbrBackground = %COLOR_MENU +1
  wClass.lpszClassName = VARPTR(szChildClass)
  wClass.hIConSm       = %NULL

' Register the child window class
' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  CALL RegisterClassEx (wClass)

' Fill Menu window class structure
' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  wClass.style         = %NULL
  wClass.lpfnWndProc   = CODEPTR(pbMenuProc)
  wClass.cbWndExtra    = 0
  wClass.hIcon         = %NULL
  wClass.hbrBackground = %COLOR_MENU +1
  wClass.lpszClassName = VARPTR(szMenuClass)
  wClass.hIConSm       = %NULL

' Register the Menu window class
' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  CALL RegisterClassEx (wClass)

' Create the main window
' ~~~~~~~~~~~~~~~~~~~~~~
  hWnd = CreateWindow (szClassName, "PB Mine", %WS_OVERLAPPEDWINDOW,_
                       %CW_USEDEFAULT, %CW_USEDEFAULT, %CW_USEDEFAULT, %CW_USEDEFAULT,_
                       %NULL, %NULL, hCurrInstance, %NULL)

  CALL ShowWindow   (hWnd, nCmdShow)
  CALL UpdateWindow (hWnd)

' Main message loop of program
' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  WHILE GetMessage(Msg, %NULL, 0, 0)
    TranslateMessage Msg
    DispatchMessage Msg
  WEND

' Clean up and assign return value
' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  FUNCTION = Msg.wParam

END FUNCTION
'
'------------------------------------
'====================================
FUNCTION MainWndProc(BYVAL hWnd    AS DWORD,_
                     BYVAL wMsg    AS DWORD,_
                     BYVAL wParam  AS DWORD,_
                     BYVAL lParam  AS DWORD) EXPORT AS LONG

  STATIC hInstance   AS LONG, _
         hWndMnu     AS DWORD, _
         hLib        AS DWORD, _
         winReSized  AS LONG, _
         getSizes    AS LONG, _
         xtraHt      AS LONG, _
         xtraWdth    AS LONG, _
         allowCheat  AS LONG, _   ' Flag indicating cheat mode is still permitted.
         cheating    AS LONG, _   ' Temporary Flag indicating that cheat mode is in use.
         szINIpath   AS ASCIIZ*%MAX_PATH

  LOCAL tValue        AS LONG, _
        rowPtr        AS LONG, _
        colPtr        AS LONG, _
        dlgReturnCode AS LONG, _
        cxClient      AS LONG, _
        cyClient      AS LONG, _
        noBlanks      AS LONG, _
        rectC         AS RECT, _
        rectW         AS RECT, _
        tileInfo      AS TileInfoTYPE, _
        ncm           AS NONCLIENTMETRICS

  SELECT CASE wMsg
    CASE %WM_CREATE
'   ~~~~~~~~~~~~~~~
' Get the height of the menu bar
      ncm.cbSize = SIZEOF(ncm)
      CALL SystemParametersInfo(BYVAL %SPI_GETNONCLIENTMETRICS, _
                                BYVAL SIZEOF(ncm), ncm, BYVAL 0)
      gmenuHt = MAX&(ncm.imenuHeight, 16)  ' Minimum height for best display
      tValue  = GetSystemMetrics(%SM_CXEDGE)
      gmenuHt = gmenuHt + (tValue * 2) + 1 ' Allow extra height for timer window border
      hInstance = GetWindowLong(hWnd, %GWL_HINSTANCE)

      DIM Kolors(1:8) ' Initiate Colors for numbers 1-8
      Kolors(1) = RGB(0,0,255)     ' Blue
      Kolors(2) = RGB(0,128,0)     ' Green
      Kolors(3) = RGB(255,0,0)     ' Red
      Kolors(4) = RGB(0,64,128)    ' Dark Blue
      Kolors(5) = RGB(156,0,0)     ' Dark Red
      Kolors(6) = RGB(64,128,128)  ' Cyan
      Kolors(7) = RGB(0,0,0)       ' Black
      Kolors(8) = RGB(128,128,128) ' Gray

' Load RichEdit's dll
      hLib = LoadLibrary("RICHED32.DLL")
' Create menu Window
      hWndMnu = CreateWindow(szmenuClass, BYVAL %NULL, _
                 %WS_CHILDWINDOW OR %WS_VISIBLE, 0,0,0,0, _
                 hWnd, %NULL, hInstance, %NULL)
' Read/Set Up game defaults
      CALL GetModuleFileName(0,szINIPath,%MAX_PATH)
      szINIPath = LEFT$(szINIPath, INSTR(-1,szINIPath, ANY "\:") ) + "PBMine.INI"
      CALL ReadINI(szINIPath)
      REDIM gGameBoard(0:gGameInfo.gameRows+1, 0:gGameInfo.gameCols+1)
      REDIM gCheatInfo(0:gGameInfo.noTiles) ' Will Contain a list of blank items

      CALL CreateGameBoard
      CALL CreateTiles(hWnd, hInstance)
      CALL AssignTileValues
      CALL MakeCheatList

      getSizes         = 0   ' A one-time flag for processing initial information.
      ghMemDCpic       = 0
      ghBitmapPic      = 0
      ghOldBitmapPic   = 0
      cheating         = %FALSE
      gTimerRunning    = %FALSE
      CALL PostMessage(hWnd, %WM_EXITSIZEMOVE, BYVAL %CENTERWINDOW, BYVAL %NULL)
      FUNCTION = 0

' A set of resizing messages/routines follows:
    CASE %WM_EXITSIZEMOVE
'   ~~~~~~~~~~~~~~~~~~~~~
      CALL GetClientRect (hWnd, RectC) ' get current size of Client Area
      winReSized = %TRUE
      CALL PostMessage(hWnd, %WM_SIZE, wParam, MAKDWD(rectC.nRight, rectC.nBottom) )
      FUNCTION = 0

    CASE %WM_SYSCOMMAND
'   ~~~~~~~~~~~~~~~~~~~
      IF wParam = %SC_ICON AND gGameInfo.gameEnd = 0 THEN
        CALL PostMessage(hWndMnu, %USR_PAUSE_TIMER, BYVAL %NULL, BYVAL %NULL)
      END IF
      IF wParam = %SC_MAXIMIZE OR wParam = %SC_RESTORE THEN
        winReSized = %TRUE
        IF gGameInfo.gameEnd = 0 THEN
          CALL PostMessage(hWndMnu, %USR_RESTART_TIMER, BYVAL %NULL, BYVAL %NULL)
        END IF
      END IF
      FUNCTION = DefWindowProc(hWnd, wMsg, wParam, lParam)  'performs default processing of message

    CASE %WM_SIZE
'   ~~~~~~~~~~~~~
      IF wParam = %SIZE_MINIMIZED OR winReSized = %FALSE THEN
        FUNCTION = 0 : EXIT SELECT
      END IF
      winReSized = %FALSE          ' Reset flag for resizing
      cxClient    = LOWRD(lParam)  ' width of client area
      cyClient    = HIWRD(lParam)  ' height of client area
' Initially find the extra size of the parent window borders and title bar.
      IF getSizes = 0 THEN  ' Calculated once as border sizes are assumed constant
        winResized = %TRUE  ' Flag for repainting child windows
        CALL GetWindowRect (hWnd, RectW)
        xtraHt   = rectW.nBottom - rectW.nTop  - cyClient
        xtraWdth = rectW.nRight  - rectW.nLeft - cxClient
' resize parent window if required
        IF gGameInfo.buttonSize < 1000 THEN  ' Values have been saved previously
          cxClient = gGameInfo.buttonSize * gGameInfo.gameCols
          cyClient = gGameInfo.buttonSize * gGameInfo.gameRows
          CALL MoveWindow (hWnd, rectW.nRight, rectW.nTop, _  ' Desktop horizontal, vertical position
                           cxClient + xtraWdth, cyClient + xtraHt + gmenuHt, %TRUE)
        END IF
        CALL GetWindowRect (hWnd, RectW) ' re-read alterations
        CALL GetClientRect (hWnd, RectC) ' re-read current size
        cyClient = rectC.nBottom         ' Reset reported sizes
        cxClient = rectC.nRight          '  "       "       "
      END IF
' Shall we resize Tiles?
      tValue = MAX(16, MIN(cxClient \ gGameInfo.gameCols, _
                          (cyClient-gmenuHt) \ (gGameInfo.gameRows)) ) ' tile minimum size: 16 pixels
      IF tValue <> gGameInfo.buttonSize OR getSizes = 0 THEN
        winResized = %TRUE  ' Flag for repainting child windows
        getSizes = 1 ' Set the once-only flag.
        gGameInfo.buttonSize = tValue
' A resized tile requires resized bitmaps.
        CALL MakeBitMap(hWnd, gGameInfo.buttonSize-2, gGameInfo.ButtonSize-2)
        FOR rowPtr = 1 TO gGameInfo.gameRows
          FOR colPtr = 1 TO gGameInfo.gameCols
            CALL MoveWindow (gGameBoard(rowPtr, colPtr).bHandle, _      ' Child Window handle
                            (colPtr-1)*gGameInfo.buttonSize, _          ' horizontal position
                            (rowPtr-1)*gGameInfo.buttonSize + gmenuHt, _ ' vertical position
                            gGameInfo.buttonSize, gGameInfo.buttonSize, %FALSE) ' don't repaint
          NEXT colPtr
        NEXT rowPtr
' The menu bar requires a certain minimum width to display everything.
        cxClient = MAX&(gMinimumwidth, gGameInfo.buttonSize * gGameInfo.gameCols)
' resize menu bar to fit main window
        CALL MoveWindow (hWndMnu, 0, 0, cxClient, gmenuHt, %TRUE)
' Write new values as defaults
        CALL WriteINI(szINIPath)
      END IF
'-----------
' Set the expected Parent Window size
      cxClient = MAX&(gMinimumwidth, gGameInfo.buttonSize * gGameInfo.gameCols) + xtraWdth
      cyClient = gGameInfo.buttonSize * gGameInfo.gameRows + xtraHt + gmenuHt
' Re-size Parent Window to accomodate - Center if This is the first session
      IF wParam = %CENTERWINDOW THEN
        CALL GetWindowRect(GetDesktopWindow, rectC)
        RectW.nLeft = MAX&( (RectC.nRight - cxClient)/2 , 0)
        RectW.nTop  = MAX&( (RectC.nBottom - cyClient)/2, 0)
      ELSE
        CALL GetWindowRect (hWnd, RectW)
      END IF
      IF wParam <> %CENTERWINDOW OR rectW.nRight <> cxClient OR rectW.nBottom <> cyClient THEN
        winResized = %TRUE
        CALL MoveWindow (hWnd, _         ' Parent Window handle
                         RectW.nLeft, RectW.nTop, _  ' Desktop horizontal/vertical position
                         cxClient, cyClient, %TRUE) 'width, height, repaint flag
' Resize Menu if necessary
        CALL GetClientRect (hWnd, RectC) ' re-read current size
        IF RectC.nRight > (cxClient - xtraWdth) THEN
          CALL MoveWindow (hWndMnu, 0, 0, rectC.nRight, gmenuHt, %TRUE)
        END IF
      END IF
      IF winResized = %TRUE THEN ' Flag for repainting child windows
        winResized = %FALSE
        CALL InvalidateRect (hWnd, BYVAL %NULL, %TRUE)  ' repaint Child Windows
      END IF
      FUNCTION = 0

    CASE %EXPOSEBLANK   ' Message from exposed button triggers exposing other buttons.
'   ~~~~~~~~~~~~~~~~~~
      IF gTimerRunning = %FALSE THEN
        CALL PostMessage(hWndMnu, %USR_START_TIMER, BYVAL %NULL, BYVAL %NULL)
      END IF
      IF wParam > 1000 OR (allowCheat >=0 AND cheating = %FALSE) THEN
        IF allowCheat >=0 THEN
          allowCheat = -1
          CALL PostMessage (hWndMnu, %WM_COMMAND, %USR_CHEAT, %MF_GRAYED)
        END IF
        FUNCTION = 0
        EXIT FUNCTION
      END IF

      IF cheating = %TRUE THEN
        cheating = %FALSE
      END IF
'     rowPtr = wParam : colPtr = lParam
      CALL ExposeToEdge(wParam, lParam, allowCheat)  ' Recursively expose edges
      IF IsWinner = 1 THEN
        CALL PostMessage(hWnd, %HITBOMB, rowPtr, colPtr) ' Simulate game end
        FUNCTION = 0
        EXIT FUNCTION
      END IF
      FUNCTION = 0

    CASE %HITBOMB
'   ~~~~~~~~~~~~~
      CALL PostMessage(hWndMnu, %USR_PAUSE_TIMER, BYVAL %NULL, BYVAL %NULL)
      FOR rowPtr = 1 TO gGameInfo.gameRows
        FOR colPtr = 1 TO gGameInfo.gameCols
' Get tile's current status and display flags
          CALL GetWindowValues(gGameBoard(rowPtr,colPtr).bHandle, tileInfo)
' Set Undo Status
          tileInfo.bUdisplay = tileInfo.bDisplay
' Change display flags for bombs
          IF tileInfo.bStatus = 10 THEN         ' Bomb here
            IF gGameInfo.gameEnd = 2 THEN
' If user won, then flag all bombs as correctly identified!
              tileInfo.bDisplay = 1
            ELSEIF tileInfo.bDisplay = 6 THEN ' This one was hit!
              tileInfo.bUdisplay = 1
            ELSEIF tileInfo.bDisplay <> 1 THEN ' Not marked, or marked incorrectly!
' Display Bomb on standard background
              tileInfo.bDisplay = 5
            END IF
          ELSEIF tileInfo.bDisplay = 1 THEN ' Flag placed in Wrong spot!
            tileInfo.bDisplay = 3           ' Display X through flag
          END IF
' Display any changes on tile!
          CALL SetWindowValues(gGameBoard(rowPtr,colPtr).bHandle, tileInfo)
        NEXT colPtr
      NEXT rowPtr
      IF gGameInfo.gameEnd <> 2 THEN gGameInfo.gameEnd = 1
      CALL InvalidateRect (hWnd, BYVAL %NULL, %TRUE)  ' repaint!
      IF gGameInfo.GameEnd = 2 THEN
        CALL PostMessage (hWnd, %WM_COMMAND, %WINNER, 0)  ' Congratulations
      END IF
      FUNCTION = 0

    CASE %WM_KEYUP  ' Translate Key codes
'   ~~~~~~~~~~~~~~
      SELECT CASE (wParam)
        CASE %VK_F1  ' Help
          CALL PostMessage (hWnd, %WM_COMMAND, %IDM_HELP, 0)

        CASE %VK_F2  ' New Game
          CALL PostMessage (hWnd, %WM_COMMAND, %IDM_NEW, 0)

        CASE %VK_F3  ' Undo
          CALL PostMessage (hWnd, %WM_COMMAND, %IDM_UNDO, 0)

        CASE %VK_F6  ' Cheat
          CALL PostMessage (hWnd, %WM_COMMAND, %IDM_CHEAT, 0)

        CASE %VK_S ' Settings hot key
          CALL PostMessage(hWndMnu, wMsg, wParam, lParam)

      END SELECT
      FUNCTION = 0

    CASE %WM_COMMAND
'   ~~~~~~~~~~~~~~~~
      SELECT CASE LOWRD(wParam)
        CASE %IDM_NEW  ' Create New Game Board
'       ~~~~~~~~~~~~~
          gGameInfo.noExposed = 0
          gGameInfo.BombCount = 0
          gGameInfo.FlagCount = 0
          gGameInfo.gameEnd   = 0
          allowCheat          = 0
          CALL PostMessage (hWndMnu, %WM_COMMAND, %USR_CHEAT, %MF_ENABLED)
          CALL CreateGameBoard
          CALL AssignTileValues
          CALL MakeCheatList
          CALL PostMessage(hWnd, %WM_EXITSIZEMOVE, %NULL, %NULL )
          CALL PostMessage(hWndMnu, %USR_ZERO_TIMER, BYVAL %NULL, BYVAL %NULL)
          FUNCTION = 0

        CASE %IDM_CHEAT
'       ~~~~~~~~~~~~~~~
          IF allowCheat < 0 OR gCheatInfo(0) = 0 THEN ' Check to see if cheat is permitted.
            BEEP
            EXIT SELECT
          END IF
' If we have already used the cheat list, there may be blanks. Delete them.
          IF allowCheat > 0 THEN CALL CompactCheatList
' Select a tile to expose
          RANDOMIZE
          noBlanks = LOWRD(gCheatInfo(0))
          IF noBlanks < 1 THEN
            noBlanks = noBlanks + HIWRD(gCheatInfo(0))
          END IF
          DO
            tValue = RND(1, noBlanks)
            rowPtr = LOWRD(gCheatInfo(tValue))
            colPtr = HIWRD(gCheatInfo(tValue))
            IF rowPtr > 0 THEN
              gCheatInfo(tValue) = 0
              EXIT LOOP
            END IF
          LOOP
          cheating = %TRUE  ' Set temporary flag. Currently cheating
          INCR allowCheat
          IF allowCheat >= %CheatLevel THEN
            allowCheat = -1
            CALL PostMessage (hWndMnu, %WM_COMMAND, %USR_CHEAT, %MF_GRAYED)
          END IF
          CALL PostMessage(hWnd, %EXPOSEBLANK, rowPtr, colPtr)
          FUNCTION = 0

        CASE %IDM_UNDO
'       ~~~~~~~~~~~~~~
          IF gGameInfo.gameEnd <> 1 THEN EXIT SELECT  ' Only undoes from a bomb hit!
          CALL PostMessage(hWndMnu, %USR_RESTART_TIMER, BYVAL %NULL, BYVAL %NULL)
          INCR gGameInfo.bombCount  ' Add new hit to the bomb count
          INCR gGameInfo.flagCount  ' Number of correctly flagged tiles
' Reset the Game End for further play
          gGameInfo.gameEnd = 0
          FOR rowPtr = 1 TO gGameInfo.gameRows
            FOR colPtr = 1 TO gGameInfo.gameCols
              CALL PostMessage (gGameBoard(rowPtr, colPtr).bHandle, %UNDO, 0, 0 )
            NEXT colPtr
          NEXT rowPtr
          CALL InvalidateRect (hWnd, BYVAL %NULL, %TRUE)  ' repaint Child Windows
' Check for flagging having gotten the last bomb
          IF IsWinner = 1 THEN
            CALL MessageBox (hWnd, "A Win by default doesn't really count!", "PBMine", _
                 %MB_ICONINFORMATION OR %MB_OK)
          END IF
          FUNCTION = 0

        CASE %WINNER
'       ~~~~~~~~~~~~
          CALL PostMessage(hWndMnu, %USR_PAUSE_TIMER, BYVAL %NULL, BYVAL %NULL)
          gTimerRunning = %FALSE
          CALL HelpBoxMaker(BYVAL hWnd, BYVAL -1, BYVAL -1, BYVAL 120, BYVAL 70, "Winner", "RTFWINNER" )
          FUNCTION = 0

        CASE %IDM_OPTIONS
'       ~~~~~~~~~~~~~~~~~
          IF gTimerRunning = %TRUE THEN
            CALL PostMessage(hWndMnu, %USR_PAUSE_TIMER, BYVAL %NULL, BYVAL %NULL)
          END IF
          dlgReturnCode = DialogBox (hInstance, BYVAL(CDWD(%dlgOptions)), _
                          hWnd, CODEPTR(OptionsProc))
          IF dlgReturnCode = 1 THEN
' Write new values as defaults
            CALL WriteINI(szINIPath)
            CALL DestroyTiles()
            gGameInfo = gNewGameInfo
            REDIM gGameBoard(0:gGameInfo.gameRows+1, 0:gGameInfo.gameCols+1)
            CALL CreateGameBoard
            CALL CreateTiles(hWnd, hInstance)
            CALL AssignTileValues
            CALL MakeCheatList
            allowCheat = 0
            CALL PostMessage(hWndMnu, %USR_ZERO_TIMER, BYVAL %NULL, BYVAL %NULL)
' Force Drawing of the New Board
            CALL PostMessage(hWnd, %WM_EXITSIZEMOVE, %NULL, %NULL )
          ELSE
            IF gTimerRunning = %TRUE AND gGameInfo.gameEnd = 0 THEN  ' Watch out for bomb hit.
              CALL PostMessage(hWndMnu, %USR_RESTART_TIMER, BYVAL %NULL, BYVAL %NULL)
            END IF
          END IF
          FUNCTION = 0

        CASE %IDM_EXIT
'       ~~~~~~~~~~~~~~
          IF hLib THEN
            Freelibrary hLib
            hLib = 0
          END IF
          IF hWndMnu THEN
            CALL DestroyWindow(hWndMnu)
            hWndMnu = 0
          END IF
          CALL DestroyTiles()
          IF ghMemDCpic <> 0 THEN
            CALL DeleteObject (SelectObject (ghMemDCpic, ghOldBitmapPic ))
            CALL DeleteObject (ghMemDCpic)
            ghMemDCpic = 0
          END IF
          IF ghMemDCnum <> 0 THEN
            CALL DeleteObject (SelectObject (ghMemDCnum, ghOldBitmapNum ))
            CALL DeleteObject (ghMemDCnum)
            ghMemDCnum = 0
          END IF
          IF ghMemDCDisp <> 0 THEN
            CALL DeleteObject (SelectObject (ghMemDCDisp, ghOldBitmapDisp ))
            CALL DeleteObject (ghMemDCDisp)
            ghMemDCDisp = 0
          END IF
          CALL PostQuitMessage(0)
          FUNCTION = 0

        CASE %IDM_ABOUT
'       ~~~~~~~~~~~~~~~
          IF hLib = 0 THEN
            FUNCTION = 0
            EXIT FUNCTION
          END IF
          IF gTimerRunning = %TRUE THEN
            CALL PostMessage(hWndMnu, %USR_PAUSE_TIMER, BYVAL %NULL, BYVAL %NULL)
          END IF
          CALL HelpBoxMaker(BYVAL hWnd, BYVAL -1, BYVAL -1, BYVAL 320, BYVAL 208, "About", "RTFABOUT" )
          IF gTimerRunning = %TRUE THEN
            CALL PostMessage(hWndMnu, %USR_RESTART_TIMER, BYVAL %NULL, BYVAL %NULL)
          END IF
          FUNCTION = 0

        CASE %IDM_HELP
'       ~~~~~~~~~~~~~~
          IF hLib = 0 THEN
            FUNCTION = 0
            EXIT FUNCTION
          END IF
          IF gTimerRunning = %TRUE THEN
            CALL PostMessage(hWndMnu, %USR_PAUSE_TIMER, BYVAL %NULL, BYVAL %NULL)
          END IF
          CALL HelpBoxMaker(BYVAL hWnd, BYVAL -1, BYVAL -1, BYVAL 360, BYVAL 280, "Help", "RTFINSTRUCT" )
          IF gTimerRunning = %TRUE THEN
            CALL PostMessage(hWndMnu, %USR_RESTART_TIMER, BYVAL %NULL, BYVAL %NULL)
          END IF
          FUNCTION = 0

      END SELECT

    CASE %WM_DESTROY
'   ~~~~~~~~~~~~~~~
      IF hLib THEN
        Freelibrary hLib
        hLib = 0
      END IF
      CALL DestroyTiles()
      IF hWndMnu THEN
        CALL DestroyWindow(hWndMnu)
        hWndMnu = 0
      END IF
      IF ghMemDCpic <> 0 THEN
        CALL DeleteObject (SelectObject (ghMemDCpic, ghOldBitmapPic ))
        CALL DeleteObject (ghMemDCpic)
        ghMemDCpic = 0
      END IF
      IF ghMemDCnum <> 0 THEN
        CALL DeleteObject (SelectObject (ghMemDCnum, ghOldBitmapNum ))
        CALL DeleteObject (ghMemDCnum)
        ghMemDCnum = 0
      END IF
      IF ghMemDCDisp <> 0 THEN
        CALL DeleteObject (SelectObject (ghMemDCDisp, ghOldBitmapDisp ))
        CALL DeleteObject (ghMemDCDisp)
        ghMemDCDisp = 0
      END IF
      CALL PostQuitMessage(0)
      FUNCTION = 0

    CASE ELSE
'   ~~~~~~~~~
      FUNCTION = DefWindowProc(hWnd, wMsg, wParam, lParam)

  END SELECT

END FUNCTION
'
' Insert a formatted RTF string into Rich Edit
'------------------------------------------------------------------------------
SUB RichEditSetString (BYVAL hRichEdit AS DWORD)
  LOCAL eStream AS EDITSTREAM

  gPos                = 1             'position in text to start from
  gPtr                = STRPTR(gTxt)  'pointer to global text buffer
  eStream.pfnCallback = CODEPTR(RichEditStreamInString) 'pointer to RichEdit callback procedure
  SendMessage hRichEdit, %EM_STREAMIN, %SF_RTF, VARPTR(eStream) 'stream in text

END SUB

'
' Rich Edit stream in callback - for streaming in string contents
'------------------------------------------------------------------------------
FUNCTION RichEditStreamInString (BYVAL dwCookie AS DWORD, BYVAL pbBuff AS BYTE PTR, _
                                 BYVAL cb AS LONG, pcb AS LONG) AS DWORD

  pcb = cb               'number of bytes to copy
  IF pcb > 0 THEN        'copy block from global string directly into Richedit's buffer.
     CopyMemory pbBuff, (gPtr + gPos - 1), pcb
     gPos = gPos + pcb   'incr pos for next callback position.
  END IF

END FUNCTION

'
'------------------------------------
'====================================
SUB HelpBoxMaker(BYVAL hParent AS DWORD, BYVAL xPos AS LONG, BYVAL yPos AS LONG, _
                 BYVAL xWidth AS LONG, BYVAL yHeight AS LONG, BYVAL inTitle AS STRING, _
                 BYVAL inText AS STRING)
  LOCAL  hDlg  AS DWORD, _
         lRes  AS DWORD, _
         lRes2 AS DWORD, _
         lRes3 AS DWORD
  STATIC hModInst AS DWORD

  hModInst = GetModuleHandle("")                                    'Get Module Handle
  lRes     = FindResource(hModInst, BYCOPY inText, BYVAL %RT_RCDATA)'look for RTF text in resource
  IF lRes  THEN lRes2 = LoadResource(hModInst, lRes)                'Load it
  IF lRes2 THEN lRes3 = LockResource(lRes2)                         'Lock it
  IF lRes3 THEN gTxt = PEEK$(lRes3, SizeofResource(hModInst, lRes)) 'Put RTF into a global string

  DIALOG NEW hParent, inTitle, xPos, yPos, xWidth, yHeight, %WS_CAPTION OR %WS_SYSMENU, 0 TO hDlg

  CONTROL ADD "RichEdit", hDlg, %ID_RICHEDIT, "", 5, 5, xWidth - 9, yHeight - 32, _
              %WS_CHILD OR %WS_VISIBLE OR %WS_TABSTOP OR %WS_VSCROLL OR _
              %ES_AUTOVSCROLL OR %ES_MULTILINE OR %ES_NOHIDESEL OR _
              %ES_READONLY OR %ES_SAVESEL, %WS_EX_CLIENTEDGE

  IF LEN(gTxt) THEN RichEditSetString GetDlgItem(hDlg, %ID_RICHEDIT)

  CONTROL ADD BUTTON, hDlg, %IDCANCEL, "&Close", xWidth - 56, yHeight - 20, 50, 14

  DIALOG SHOW MODAL hDlg, CALL HelpBoxProc

END SUB
'
'------------------------------------
'====================================
CALLBACK FUNCTION HelpBoxProc
  SELECT CASE CBMSG
     CASE %WM_INITDIALOG

     CASE %WM_COMMAND
        SELECT CASE CBCTL
            CASE %IDCANCEL 'respond to ESC key (or button with %IDCANCEL id)
              IF CBCTLMSG = %BN_CLICKED OR CBCTLMSG = 1 THEN DIALOG END CBHNDL
        END SELECT

  END SELECT
END FUNCTION
'
'------------------------------------
'====================================
FUNCTION pbMenuProc(BYVAL hWnd    AS DWORD,_
                    BYVAL wMsg    AS DWORD,_
                    BYVAL wParam  AS DWORD,_
                    BYVAL lParam  AS DWORD) EXPORT AS LONG

  STATIC hWndButton  AS LONG, _
         hWndTime    AS LONG, _
         hMenu       AS LONG, _
         txtSize     AS sizeL, _  ' Button dimensions
         stSize      AS sizeL, _  ' Static Label dimensions
         timerSize   AS sizeL, _  ' timer number dimensions (within Static Label)
         tBorder     AS sizeL, _  ' border dimensions for drawEdge()
         hTimer      AS LONG, _
         displayTime AS LONG, _
         startTime   AS QUAD, _
         sleepTime   AS QUAD, _
         sleepTimeS  AS QUAD, _
         elapsedTime AS QUAD, _
         ncm         AS NONCLIENTMETRICS

  LOCAL hdc        AS LONG, _
        pdrawItem  AS DRAWITEMSTRUCT PTR, _
        oText      AS STRING, _
        tValue     AS LONG, _
        rectC      AS RECT, _
        hFont      AS LONG, _
        st         AS SYSTEMTIME, _
        ft         AS FILETIME, _
        QT1        AS QuadFILETIME, _
        currentTime AS QUAD, _
        calcTime   AS QUAD

  SELECT CASE wMsg
    CASE %WM_CREATE
'   ~~~~~~~~~~~~~~~
      tBorder.cx = GetSystemMetrics(%SM_CXEDGE)
      tBorder.cy = GetSystemMetrics(%SM_CYEDGE)
' Create Owner Drawn Button for menu.
      oText = "Settings"
      hdc = GetDC(hWnd)
      ncm.cbSize = SIZEOF(ncm)
      CALL SystemParametersInfo(BYVAL %SPI_GETNONCLIENTMETRICS, _
                                BYVAL SIZEOF(ncm), ncm, BYVAL 0)

      hFont = CreateFontIndirect(ncm.lfMenuFont)
      CALL SelectObject(hDC, hFont)
      CALL GetTextExtentPoint32( hDC, BYCOPY oText, LEN(oText), txtSize )
      txtSize.cx = txtSize.cx + tBorder.cx * 2
      CALL DeleteObject(hFont)
      CALL ReleaseDC(hWnd, hDC)
      hWndButton = CreateWindow("BUTTON", BYVAL %NULL, _
          %WS_CHILDWINDOW OR %WS_VISIBLE OR %BS_OWNERDRAW, _
          0,0,txtSize.cx, gmenuHt - 1, _
          hWnd, %IDM_HDRBUTTON, GetWindowLong(hWnd, %GWL_HINSTANCE), %NULL)
' Create Static Label to display timer information. Final dimensions made later.
      hWndTime = CreateWindow("STATIC", BYVAL %NULL, _
          %WS_CHILDWINDOW OR %WS_VISIBLE OR %SS_OWNERDRAW, _
          0,0,0,0, _
          hWnd, %IDM_TMRBUTTON, GetWindowLong(hWnd, %GWL_HINSTANCE), %NULL)
' Calculate dimensions for Static Label and timer numeric display
      stSize.cy    = gmenuHt - 1
      timerSize.cy = stSize.cy - (tBorder.cy * 2)
      timerSize.cx = timerSize.cy * .7
      stSize.cx    = (timerSize.cx * 3) + (tBorder.cx * 2)
' Fill Global with minimum width permitted in order to display the button and the timer window.
      gMinimumWidth = stSize.cx + txtSize.cx
' Create LED Number Bitmap
      CALL MakeLEDBitMap(hWnd, timerSize.cx, timerSize.cy)
' Create Bitmap Timer Display to be held by a Static Label
      hDC             = GetDC(hWnd)
      ghMemDCDisp     = CreateCompatibleDC(%NULL)
      ghBitmapDisp    = CreateCompatibleBitmap(hDC, timerSize.cx*3, timerSize.cy)
      ghOldBitmapDisp = SelectObject(ghMemDCDisp, ghBitmapDisp)
      CALL ReleaseDC(hWnd, hDC)

' Create Popup Menu
      hMenu = CreatePopupMenu ()
      CALL AppendMenu (hMenu, %MF_STRING,    %IDM_NEW,    "&New Game F2")
      CALL AppendMenu (hMenu, %MF_STRING,    %IDM_UNDO,   "&Undo     F3")
      CALL AppendMenu (hMenu, %MF_STRING,    %IDM_OPTIONS,"&Options ...")
      CALL AppendMenu (hMenu, %MF_SEPARATOR, 0,  "")
      CALL AppendMenu (hMenu, %MF_STRING,    %IDM_CHEAT, "&Expose Tile F6")
      CALL AppendMenu (hMenu, %MF_SEPARATOR, 0,  "")
      CALL AppendMenu (hMenu, %MF_STRING,    %IDM_EXIT,   "&Exit")
      CALL AppendMenu (hMenu, %MF_SEPARATOR, 0,  "")
      CALL AppendMenu (hMenu, %MF_STRING, %IDM_HELP, "&Help")
      CALL AppendMenu (hMenu, %MF_STRING, %IDM_ABOUT, "&About")

      CALL PostMessage(hWnd, %USR_ZERO_TIMER, BYVAL %NULL, BYVAL %NULL)
      FUNCTION = 0

    CASE %WM_SIZE
'   ~~~~~~~~~~~~~
      CALL GetClientRect(hWnd, rectC)
      CALL MoveWindow(hWndTime, rectC.nRight - stSize.cx, 0, _
                      stSize.cx, stSize.cy, %TRUE )
      FUNCTION = 0

    CASE %WM_DRAWITEM ' Owner-Draw Timer Label & Button used to simulate main menu Item - Settings
'   ~~~~~~~~~~~~~~~~~
      pDrawItem = lParam
      IF @pDrawItem.CtlID = %IDM_HDRBUTTON AND @pDrawItem.itemAction = %ODA_DRAWENTIRE THEN
        oText = "&Settings"
        hFont = CreateFontIndirect(ncm.lfMenuFont)
        CALL SelectObject(@pDrawItem.hDC, hFont)
        CALL DrawText(@pDrawItem.hDC, BYCOPY(oText), LEN(oText), @pDrawItem.rcItem, _
                      %DT_CENTER OR %DT_SINGLELINE)
        CALL DeleteObject(hFont)

      ELSEIF @pDrawItem.CtlID = %IDM_TMRBUTTON THEN
        CALL DrawEdge(@pDrawItem.hDC, @pDrawItem.rcItem, %EDGE_SUNKEN, %BF_RECT)
        CALL BitBlt(@pDrawItem.hDC, @pDrawItem.rcItem.nLeft + tBorder.cx, _
                    @pDrawItem.rcItem.nTop + tBorder.cy, timerSize.cx*3, timerSize.cy, _
                    ghMemDCDisp, 0, 0, %SRCCOPY)
      END IF
      FUNCTION = 0
      EXIT FUNCTION

    CASE %WM_KEYUP  ' Translate Key codes
'   ~~~~~~~~~~~~~~
      SELECT CASE (wParam)
        CASE %VK_F1  ' Help
          CALL PostMessage (GetParent(hWnd), %WM_COMMAND, %IDM_HELP, 0)

        CASE %VK_F2  ' New Game
          CALL PostMessage (GetParent(hWnd), %WM_COMMAND, %IDM_NEW, 0)

        CASE %VK_F3  ' Undo
          CALL PostMessage (GetParent(hWnd), %WM_COMMAND, %IDM_UNDO, 0)

        CASE %VK_F6  ' Cheat
          CALL PostMessage (GetParent(hWnd), %WM_COMMAND, %IDM_CHEAT, 0)

        CASE %VK_S ' Settings hot key. Simulate a button press.
          tValue = MAKLNG(0,%BN_CLICKED)
          CALL PostMessage(hWnd, %WM_COMMAND, tValue, hWndButton)

      END SELECT
      FUNCTION = 0

    CASE %USR_START_TIMER ' Set timer going
'   ~~~~~~~~~~~~~~~~~~~~~
      IF hTimer <> 0 THEN
        CALL KillTimer(hWnd, hTimer)
        hTimer = 0
      END IF
      CALL GetLocalTime(st)
      CALL SystemTimeToFileTime(st, ft)
      QT1.dwDateTime = ft
      startTime   = QT1.qdTime
      displayTime = 0
      sleepTimeS  = 0
      sleepTime   = 0
      elapsedtime = 0
      hTimer = SetTimer(hWnd, 1, %TIMER_WAIT_VALUE, BYVAL %NULL)
      gTimerRunning = %TRUE
' BitBlt '000' to initial Bitmap
      CALL PaintLED(BYVAL displayTime, timerSize.cx, timerSize.cy)
      CALL InvalidateRect(hWndTime, BYVAL %NULL, 1)
      FUNCTION = 0

    CASE %USR_ZERO_TIMER ' Zero Timer: Reset values
'   ~~~~~~~~~~~~~~~~~~~~
      IF hTimer <> 0 THEN
        CALL KillTimer(hWnd, hTimer)
        hTimer = 0
      END IF
      startTime   = 0
      displayTime = 0
      sleepTimeS  = 0
      sleepTime   = 0
      elapsedtime = 0
      gTimerRunning = %FALSE
' BitBlt '000' to initial Bitmap
      CALL PaintLED(BYVAL displayTime, timerSize.cx, timerSize.cy)
      CALL InvalidateRect(hWndTime, BYVAL %NULL, 1)
      FUNCTION = 0

    CASE %USR_RESTART_TIMER ' ReStart timer: going active/ recover from bomb blast (F3)
'   ~~~~~~~~~~~~~~~~~~~~~~~
      IF sleepTimeS = 0 THEN
        FUNCTION = 0
        EXIT FUNCTION
      END IF
      CALL GetLocalTime(st)
      CALL SystemTimeToFileTime(st, ft)
      QT1.dwDateTime = ft
      currentTime = QT1.qdTime
      sleepTime = sleepTime + currentTime - sleepTimeS
      hTimer = SetTimer(hWnd, 1, %TIMER_WAIT_VALUE, BYVAL %NULL)
      gTimerRunning = %TRUE
      sleepTimeS = 0
      FUNCTION = 0

    CASE %USR_PAUSE_TIMER ' Pause Timer: going iconic/bomb blast has occurred
'   ~~~~~~~~~~~~~~~~~~~~~
      IF hTimer = 0 THEN
        FUNCTION = 0
        EXIT FUNCTION
      END IF
      CALL KillTimer(hWnd, hTimer)
      CALL GetLocalTime(st)
      CALL SystemTimeToFileTime(st, ft)
      QT1.dwDateTime = ft
      sleepTimeS = QT1.qdTime
      FUNCTION = 0

    CASE %WM_TIMER ' Timer triggered. Do we update timer display?
'   ~~~~~~~~~~~~~~
      CALL GetLocalTime(st)
      CALL SystemTimeToFileTime(st, ft)
      QT1.dwDateTime = ft
      currentTime = QT1.qdTime
      calcTime = currentTime - startTime - sleepTime
      IF calcTime > elapsedTime + %ONE_SECOND THEN
        displaytime = (calcTime/%ONE_SECOND)
        CALL PaintLED(BYVAL displayTime, timerSize.cx, timerSize.cy)
        CALL InvalidateRect(hWndTime, BYVAL %NULL, 1)
      END IF
      FUNCTION = 0

    CASE %WM_COMMAND
'   ~~~~~~~~~~~~~~~~
      IF wParam = %USR_CHEAT THEN  ' Message passed to enable/disable Cheat Item
        CALL EnableMenuItem (hMenu, %IDM_CHEAT, lParam)
        FUNCTION = 0
        EXIT SELECT
      END IF
      IF HIWRD(wParam) = %BN_CLICKED THEN
        CALL GetWindowRect(hWndButton, rectC)
        CALL TrackPopupMenuEx(hMenu, %TPM_HORIZONTAL,_
        rectC.nLeft, rectC.nBottom, GetParent(hWnd), BYVAL %NULL)
      END IF
      CALL SetFocus(hWnd)  ' Have to take focus away from button
      FUNCTION = 0

    CASE %WM_DESTROY
'   ~~~~~~~~~~~~~~~~
      IF hTimer <> 0 THEN
        CALL KillTimer(hWnd, hTimer)
      END IF
      CALL DestroyMenu(hMenu)
      FUNCTION = 0

  END SELECT

  FUNCTION = DefWindowProc(hWnd, wMsg, wParam, lParam)

END FUNCTION
'
'------------------------------------
'====================================
SUB SetWindowValues(hWnd AS DWORD, tileInfo AS TileInfoTYPE)
' Sets Status, Display and UNDO codes
  STATIC TileConvU AS TileConvertUNION, _
         TileConvT AS TileConvertTYPE

  TileConvT.bStatus   = tileInfo.bStatus
  TileConvT.bDisplay  = tileInfo.bDisplay
  TileConvT.bUdisplay = tileInfo.bUdisplay
  TileConvU.bInfo = TileConvT
  CALL SetWindowLong(hWnd, 0, TileConvU.convert)

END SUB
'
'------------------------------------
'====================================
SUB GetWindowValues(hWnd AS DWORD, tileInfo AS TileInfoTYPE)
' Retrieves Status, Display and UNDO codes
  STATIC TileConvU AS TileConvertUNION, _
         TileConvT AS TileConvertTYPE

  TileConvU.convert  = GetWindowLong(hWnd,  0)
  tileConvT          = TileConvU.bInfo
  tileInfo.bStatus   = TileConvT.bStatus
  tileInfo.bDisplay  = TileConvT.bDisplay
  tileInfo.bUdisplay = TileConvT.bUdisplay

END SUB
'
'------------------------------------
'====================================
SUB GetCoordinates(hWnd AS DWORD, rowPtr AS LONG, colPtr AS LONG)
' Gets GameBoard Coordinates, which are used for a unique identifier for each tile.
  rowPtr = GetWindowLong(hWnd, %GWL_ID)   ' Get Tile Location
  colPtr = rowPtr
  SHIFT RIGHT colPtr, 16
  SHIFT LEFT rowPtr, 16 : SHIFT RIGHT rowPtr, 16
END SUB
'
'------------------------------------
'====================================
FUNCTION IsWinner AS LONG
' Check if the values are such that this move causes a win.
' If all the bombs are flagged, or everything that isn't a bomb is exposed...
  IF (gGameInfo.bombCount = gGameInfo.noBombs AND _
      gGameInfo.bombCount = gGameInfo.FlagCount) OR _
     (gGameInfo.noExposed = gGameInfo.noTiles - gGameInfo.noBombs) THEN
    gGameInfo.gameEnd = 2   ' A Winner!
    FUNCTION = 1
  ELSE
     FUNCTION = 0
  END IF
END FUNCTION
'
'------------------------------------
'====================================
SUB CreateGameBoard
  LOCAL i         AS LONG, _
        rowPtr    AS LONG, _
        colPtr    AS LONG, _
        rowOffset AS LONG, _
        colOffset AS LONG
' Reset Array to Zero Values
  FOR rowPtr = 1 TO gGameInfo.gameRows
    FOR colPtr = 1 TO gGameInfo.gameCols
      gGameBoard(rowPtr,colPtr).bStatus  = 0
      gGameBoard(rowPtr,colPtr).bDisplay = 0
    NEXT colPtr
  NEXT rowPtr

  RANDOMIZE
' Place Bombs
  FOR i = 1 TO gGameInfo.noBombs
    rowPtr = RND(1,gGameInfo.gameRows)
    colPtr = RND(1,gGameInfo.gameCols)
    IF gGameBoard(rowPtr, colPtr).bStatus <> 10 THEN
      gGameBoard(rowPtr, colPtr).bStatus = 10
' Add count to Nearest Neighbours
      FOR rowOffset = -1 TO 1
        FOR colOffset = -1 TO 1
          IF gGameBoard(rowPtr+rowOffset, colPtr+colOffset).bStatus < 10 THEN
            IF rowOffset <> 0 OR colOffset<>0 THEN
               INCR gGameBoard(rowPtr+rowOffset, colPtr+colOffset).bStatus
            END IF
          END IF
        NEXT colOffset
      NEXT rowOffset
    ELSE
' have to retry placing that bomb.
      DECR i
      ITERATE FOR
    END IF
  NEXT i

END SUB
'
'------------------------------------
'====================================
SUB CreateTiles(hWnd AS DWORD, hInstance AS LONG)
  STATIC rowPtr AS LONG, _
         colPtr AS LONG, _
         tileID AS LONG

  FOR rowPtr = 1 TO gGameInfo.gameRows
    FOR colPtr = 1 TO gGameInfo.gameCols
      tileID = MAKLNG(rowPtr,colPtr)
      gGameBoard(rowPtr, colPtr).bHandle = CreateWindow(szChildClass, _ 'window class name
                 BYVAL %NULL, %WS_CHILDWINDOW OR %WS_VISIBLE , 0,0,0,0, _
                 hWnd, tileID, _         'parent handle, child ID
                 hInstance, %NULL)       'program instance handle, extra parameter.
    NEXT colPtr
  NEXT rowPtr
END SUB
'
'------------------------------------
'====================================
SUB AssignTileValues
  LOCAL rowPtr AS LONG, _
        colPtr AS LONG

  FOR rowPtr = 1 TO gGameInfo.gameRows
    FOR colPtr = 1 TO gGameInfo.gameCols
      CALL PostMessage (gGameBoard(rowPtr, colPtr).bHandle, %SETVALUE, _
                        gGameBoard(rowPtr,colPtr).bStatus, _
                        gGameBoard(rowPtr,colPtr).bDisplay)
    NEXT colPtr
  NEXT rowPtr
END SUB
'
'------------------------------------
'====================================
SUB MakeCheatList
  LOCAL rowPtr AS LONG, _
        colPtr AS LONG, _
        bCount AS LONG, _  ' number of blank tiles
        nCount AS LONG     ' number of numbered tiles
' Create a list of tiles that may be harmlessly exposed
  REDIM gCheatInfo(0:gGameInfo.noTiles)
  bCount = 0
  nCount = 0
' Add the Blanks first
  FOR rowPtr = 1 TO gGameInfo.gameRows
    FOR colPtr = 1 TO gGameInfo.gameCols
      IF gGameBoard(rowPtr, colPtr).bStatus = 0 THEN
        INCR bCount
        gCheatInfo(bCount) = MAKLNG(rowPtr,colPtr)
      END IF
    NEXT colPtr
  NEXT rowPtr

' Now add the numbered tiles
  FOR rowPtr = 1 TO gGameInfo.gameRows
    FOR colPtr = 1 TO gGameInfo.gameCols
      IF gGameBoard(rowPtr, colPtr).bStatus > 0 AND _
         gGameBoard(rowPtr, colPtr).bStatus < 10 THEN
        INCR nCount
        gCheatInfo(bCount+nCount) = MAKLNG(rowPtr,colPtr)
      END IF
    NEXT colPtr
  NEXT rowPtr
  gCheatInfo(0) = MAKLNG(bCount, nCount)  ' save the number of items in each category.

END SUB
'
'------------------------------------
'====================================
SUB CompactCheatList
  LOCAL bCount    AS LONG, _
        nCount    AS LONG, _
        newBcount AS LONG, _
        newNcount AS LONG, _
        i         AS LONG, _
        tCheatInfo() AS LONG

' Create a temporary list to which to write Cheat Info
  REDIM tCheatInfo(0:gGameInfo.noTiles)
  bCount = LOWRD(gCheatInfo(0))
  newBcount = 0
  nCount = HIWRD(gCheatInfo(0))
  newNcount = 0
' Add the Blanks first
  FOR i = 1 TO bCount
    IF gCheatInfo(i)<> 0 THEN
      INCR newBcount
      tCheatInfo(newBcount) = gCheatInfo(i)
    END IF
  NEXT i
' Now add the numbered tiles
  FOR i = bCount+1 TO bCount+nCount
    IF gCheatInfo(i)<> 0 THEN
      INCR newNcount
      tCheatInfo(newBcount+newNcount) = gCheatInfo(i)
    END IF
  NEXT i
' Copy updated info to Cheat List
  gCheatInfo(0) = MAKLNG(newBcount, newNcount)
  FOR i = 1 TO newBcount+newNcount
    gCheatInfo(i) = tCheatInfo(i)
  NEXT i

END SUB
'
'------------------------------------
'====================================
SUB DestroyTiles()
  LOCAL rowPtr AS LONG, _
        colPtr AS LONG

  FOR rowPtr = 1 TO gGameInfo.gameRows
    FOR colPtr = 1 TO gGameInfo.gameCols
      CALL DestroyWindow(gGameBoard(rowPtr,colPtr).bHandle)
    NEXT colPtr
  NEXT rowPtr
END SUB
'
'------------------------------------
'====================================
FUNCTION ButtonProc(BYVAL hWnd    AS DWORD,_
                    BYVAL wMsg    AS DWORD,_
                    BYVAL wParam  AS DWORD,_
                    BYVAL lParam  AS DWORD) AS LONG

  LOCAL ps       AS PAINTSTRUCT, _
        hdc      AS LONG, _
        hPen     AS LONG, _
        hBrush   AS LONG, _
        hFont    AS LONG, _
        lFont    AS LOGFONT, _
        RectC    AS RECT, _
        rowPtr   AS LONG, _
        colPtr   AS LONG, _
        oText    AS STRING, _
        xWidth   AS LONG, _
        tileInfo AS  TileInfoTYPE

  SELECT CASE wMsg
    CASE %WM_CREATE
'   ~~~~~~~~~~~~~~~
      CALL SetWindowLong (hWnd, 0, 0) ' initialize value of User Info
      FUNCTION = 0

    CASE %SETVALUE
'   ~~~~~~~~~~~~~~
      tileInfo.bStatus   = wParam
      tileInfo.bDisplay  = lParam
      tileInfo.bUdisplay = lParam
      CALL SetWindowValues(hWnd, tileInfo) ' Set Status, Display and UNDO info
      FUNCTION = 0

    CASE %UNDO
'   ~~~~~~~~~~
      CALL GetWindowValues(hWnd, tileInfo)     ' Get Current Info
      tileInfo.bDisplay = tileInfo.bUdisplay   ' Write the UNDO info
      CALL SetWindowValues(hWnd, tileInfo)     ' Reset Information
      FUNCTION = 0

    CASE %WM_RBUTTONUP
'   ~~~~~~~~~~~~~~~~~~
      IF gGameInfo.gameEnd > 0 THEN EXIT SELECT
      CALL GetCoordinates(hWnd, rowPtr, colPtr)
      CALL GetWindowValues(hWnd, tileInfo)      ' Get Button: Contents, display status
      IF tileInfo.bDisplay = 4 THEN EXIT SELECT ' Exit if already exposed!
      IF tileInfo.bDisplay = 0 THEN
        INCR gGameInfo.flagCount       ' Counter for number of flags
        IF tileInfo.bStatus = 10 THEN
          INCR gGameInfo.bombCount     ' Counter for correct flags
        END IF
      ELSE
        IF tileInfo.bDisplay = 1 THEN
          DECR gGameInfo.flagCount
          IF tileInfo.bStatus = 10 THEN
            DECR gGameInfo.bombCount   ' In the correct place
          END IF
        END IF
      END IF
      INCR tileInfo.bDisplay          ' Toggle next flag marker: 0-2
      IF tileInfo.bDisplay > 2 THEN tileInfo.bDisplay = 0
      IF IsWinner = 1 THEN            ' Simulate game end
        CALL PostMessage(GetWindowLong(hWnd,%GWL_HWNDPARENT), %HITBOMB, rowPtr, colPtr)
      END IF
      tileInfo.bUdisplay = tileInfo.bDisplay
      CALL SetWindowValues(hWnd, tileInfo)            ' Set Status, Display, UNDO info
      CALL InvalidateRect (hWnd, BYVAL %NULL, %TRUE)  ' repaint!
      FUNCTION = 0

    CASE %WM_LBUTTONUP
'   ~~~~~~~~~~~~~~~~~~
      IF gGameInfo.gameEnd > 0 THEN EXIT FUNCTION
      CALL GetCoordinates(hWnd, rowPtr, colPtr)
      CALL GetWindowValues(hWnd, tileInfo)                   ' Get Button: Contents, display status
      tileInfo.bUdisplay = tileInfo.bDisplay
      IF tileInfo.bDisplay = 1 OR tileInfo.bDisplay = 2 THEN ' No Action - button flagged
        BEEP    ' Annoy the user.
      ELSEIF tileInfo.bStatus = 10 THEN  ' Bomb!
        tileInfo.bDisplay = 6            ' Mark square as hit
        CALL PostMessage(GetWindowLong(hWnd,%GWL_HWNDPARENT), %HITBOMB, rowPtr, colPtr)
      ELSE
        IF tileInfo.bStatus = 0 THEN
          CALL PostMessage(GetWindowLong(hWnd,%GWL_HWNDPARENT), _
               %EXPOSEBLANK, rowPtr, colPtr)
          CALL PostMessage(GetWindowLong(hWnd,%GWL_HWNDPARENT), _ ' Fake Msg
               %EXPOSEBLANK, BYVAL 9999, BYVAL 9999)       ' Turn off Cheat
        ELSEIF tileInfo.bDisplay <> 4 THEN
          CALL PostMessage(GetWindowLong(hWnd,%GWL_HWNDPARENT), _ ' Fake Msg
               %EXPOSEBLANK, BYVAL 9999, BYVAL 9999)       ' Turn off Cheat
          tileInfo.bDisplay = 4    ' Mark Button as Exposed (unless blank)
          INCR gGameInfo.noExposed  ' Keep count of number exposed for end of game calculation
          IF IsWinner = 1 THEN
            CALL PostMessage(GetWindowLong(hWnd,%GWL_HWNDPARENT), _
            %HITBOMB, rowPtr, colPtr)
          END IF
          CALL InvalidateRect (hWnd, BYVAL %NULL, %TRUE)  ' repaint!
        END IF
      END IF
      CALL SetWindowValues(hWnd, tileInfo) ' Set Information
      FUNCTION = 0

    CASE %WM_PAINT
'   ~~~~~~~~~~~~~~
      hPen = 0 : hBrush = 0 : hFont = 0
      CALL GetCoordinates(hWnd, rowPtr, colPtr)
      CALL GetWindowValues(hWnd, tileInfo)
      CALL GetClientRect(hWnd, RectC)
      hdc = BeginPaint(hWnd, ps)
      IF tileInfo.bDisplay < 4 THEN ' Unexposed tile
        hPen = CreatePen(%PS_solid,%NULL, %BLACK_PEN)
        CALL SelectObject(hDC,hPen)
        CALL DrawEdge(hdc, RectC, %EDGE_RAISED, %BF_RECT OR %BF_SOFT)
        lFont.lfFaceName       = "Arial"
        lFont.lfPitchAndFamily = 34
        lFont.lfHeight         = - gGameInfo.buttonSize * 0.8
        lFont.lfWeight         = 800
        hFont = CreateFontIndirect(lFont)
        CALL SelectObject (hdc, hFont)
        CALL SetTextColor(hdc, RGB(0,0,0) )
        CALL SetBkMode( hdc, %TRANSPARENT)

        SELECT CASE tileInfo.bDisplay
          CASE 0,2
            IF tileInfo.bDisplay = 2 THEN
               oText = "?"
               CALL DrawTextEx(hdc, BYCOPY oText, 1, RectC, %DT_CENTER OR %DT_VCENTER, BYVAL %NULL)
            END IF
          CASE 1,3
            xWidth = gGameInfo.buttonSize-2
            IF tileInfo.bDisplay = 1 THEN
              CALL BitBlt (hDC, 1, 1, xWidth, xWidth, ghMemDCpic, xWidth*2, 0, %SRCCOPY)
            ELSE
              CALL BitBlt (hDC, 1, 1, xWidth, xWidth, ghMemDCpic, xWidth*3, 0, %SRCCOPY)
            END IF
        END SELECT

      ELSE  ' Exposed Tile
        hPen = CreatePen(%PS_SOLID,%NULL, %BLACK_PEN)
        CALL SelectObject(hDC,hPen)
        CALL Rectangle(hdc, RectC.nTop, RectC.nLeft, RectC.nBottom, RectC.nRight)
        hBrush = CreateSolidBrush(RGB(255,255,255))
        CALL SelectObject(hdc,hBrush)
        CALL Rectangle(hdc, RectC.nTop, RectC.nLeft, RectC.nBottom, RectC.nRight)

        DECR RectC.nBottom
        DECR RectC.nRight
        INCR RectC.nTop
        INCR RectC.nLeft
        CALL FillRect(hdc, RectC, hBrush)

        lFont.lfFaceName       = "Arial"
        lFont.lfPitchAndFamily = 34
        lFont.lfHeight         = - gGameInfo.buttonSize * 0.8
        lFont.lfWeight         = 800
        hFont = CreateFontIndirect(lFont)
        CALL SelectObject (hdc, hFont)
        CALL SetTextColor(hdc, RGB(0,0,0) )
        CALL SetBkMode( hdc, %TRANSPARENT)
        CALL SetBkColor(hdc, RGB(192,192,192) )

        SELECT CASE tileInfo.bDisplay
          CASE 4
            IF tileInfo.bStatus <> 0 THEN
              CALL SetTextColor(hdc, Kolors(tileInfo.bStatus) )
              oText = LTRIM$(STR$(tileInfo.bStatus))
              CALL DrawTextEx(hdc, BYCOPY oText, 1, RectC, _
                   %DT_CENTER OR %DT_VCENTER, BYVAL %NULL)
            END IF
          CASE 5,6
            xWidth = gGameInfo.buttonSize-2
            IF tileInfo.bDisplay = 5 THEN
              CALL BitBlt (hDC, 1, 1, xWidth, xWidth, ghMemDCpic, 0, 0, %SRCCOPY)
            ELSE
              CALL BitBlt (hDC, 1, 1, xWidth, xWidth, ghMemDCpic, xWidth, 0, %SRCCOPY)
            END IF

        END SELECT

      END IF
      CALL ReleaseDC(hWnd, hDC)
      CALL EndPaint(hWnd, ps)

      IF hPen   > 0 THEN CALL DeleteObject(hPen)
      IF hBrush > 0 THEN CALL DeleteObject(hBrush)
      IF hFont  > 0 THEN CALL DeleteObject(hFont)

      FUNCTION = 0

    CASE ELSE
'   ~~~~~~~~~
      FUNCTION = DefWindowProc(hWnd, wMsg, wParam, lParam)

  END SELECT

END FUNCTION
'
'------------------------------------
'====================================
FUNCTION OptionsProc(BYVAL hDlg AS LONG, BYVAL wMsg AS LONG, _
                     BYVAL wParam AS LONG, BYVAL lParam AS LONG) AS LONG

  LOCAL newgameCols AS LONG, _
        newgameRows AS LONG, _
        newNoBombs  AS LONG, _
        testValue   AS LONG, _
        tString     AS ASCIIZ * 4

  SELECT CASE wMsg
    CASE %WM_INITDIALOG
'   ~~~~~~~~~~~~~~~~~~~
' Limit Edit Control Size
      CALL SendMessage(GetDlgItem(hDlg, %eWidth),  %EM_SETLIMITTEXT, 2, 0)
      CALL SendMessage(GetDlgItem(hDlg, %eHeight), %EM_SETLIMITTEXT, 2, 0)
      CALL SendMessage(GetDlgItem(hDlg, %eBombs),  %EM_SETLIMITTEXT, 3, 0)
' Fill The Edit Controls
      SetWindowText GetDlgItem(hDlg, %eWidth),   BYCOPY TRIM$(STR$(gGameInfo.gameCols))
      SetWindowText GetDlgItem(hDlg, %eHeight),  BYCOPY TRIM$(STR$(gGameInfo.gameRows))
      SetWindowText GetDlgItem(hDlg, %eBombs),   BYCOPY TRIM$(STR$(gGameInfo.noBombs))
' Select text in Edit Control receiving focus
      CALL SendMessage(GetDlgItem(hDlg, %eWidth),  %EM_SETSEL, 0, -1)
      CALL SetFocus( GetDlgItem(hDlg, %eWidth) )

    CASE %DM_GETDEFID  ' Enter Key caused windows to ask for default Push button
'   ~~~~~~~~~~~~~~~~~
      IF (GetKeyState(%VK_RETURN) AND &H8000) THEN
        testValue = GetDlgCtrlID(GetFocus())
        IF testValue = %pb_IDOK OR testValue = %IDCANCEL THEN ' allow buttons to be pushed.
          CALL PostMessage( hDlg, %WM_COMMAND, testValue, 0)
        ELSE
          SetFocus GetNextDlgTabItem( hDlg, GetFocus(), %NULL)
          CALL SendMessage(GetFocus(), %EM_SETSEL, BYVAL 0, BYVAL -1 )
          FUNCTION = 1
        END IF
      END IF

    CASE %WM_COMMAND
'   ~~~~~~~~~~~~~~~~
      SELECT CASE LOWRD(wParam)
        CASE %IDCANCEL
'       ~~~~~~~~~~~~~~
          EndDialog hDlg, 0
          FUNCTION = 1

        CASE %pb_IDOK
'       ~~~~~~~~~~~~~
' Retrieve edit control values
          CALL GetWindowText(GetDlgItem(hDlg, %eWidth), tString, 4)
          newgameCols = VAL(tString)
          CALL GetWindowText(GetDlgItem(hDlg, %eHeight), tString, 4)
          newgameRows = VAL(tString)
          CALL GetWindowText(GetDlgItem(hDlg, %eBombs), tString, 5)
          newNoBombs = VAL(tString)
' Bring values within a valid range
          gNewGameInfo.gameCols = MAX(3,MIN(newgameCols, 35))   ' 3 to 35
          gNewGameInfo.gameRows = MAX(3,MIN(newgameRows, 25))   ' 3 to 25
          gNewGameInfo.noTiles  = gNewGameInfo.gameCols * gNewGameInfo.gameRows
' Maximum 80% bombs
          gNewGameInfo.noBombs  = MAX(1, MIN(newNoBombs, 0.8*gNewGameInfo.noTiles))
          EndDialog hDlg, 1
          FUNCTION = 0
      END SELECT
    END SELECT

END FUNCTION
'
'------------------------------------
'====================================
SUB ExposeToEdge(rowPtr AS LONG, colPtr AS LONG, allowcheat AS LONG)
 ' Expose adjoining blank tiles and all their adjoining numbered tiles

  LOCAL rowOffset AS LONG, _
        colOffset AS LONG, _
        tileInfo  AS tileInfoTYPE, _
        test      AS LONG

' Get the Tile's contents and flags.
  CALL GetWindowValues(gGameBoard(rowPtr,colPtr).bHandle, tileInfo)
  IF tileInfo.bDisplay >3 THEN EXIT SUB  ' Tile is already exposed
  IF tileInfo.bDisplay = 0 THEN          ' Tile may be exposed
    INCR gGameInfo.noExposed
    tileInfo.bDisplay  = 4               ' Display contents: Blank/Number
    tileInfo.bUdisplay = 4
    CALL SetWindowValues(gGameBoard(rowPtr,colPtr).bHandle, tileInfo)
    CALL InvalidateRect (gGameBoard(rowPtr,colPtr).bHandle, BYVAL %NULL, %TRUE)
  END IF
  IF allowcheat > 0 THEN
    test = MAKLNG(rowPtr, colPtr)
    ARRAY SCAN gCheatInfo(1) FOR LOWRD(gCheatInfo(0))+HIWRD(gCheatInfo(0)), = test, TO test
    IF test > 0 THEN gCheatInfo(test) = 0
  END IF
  IF tileInfo.bStatus > 0 AND tileInfo.bStatus < 10 THEN EXIT SUB   ' Numbered tile. Stop recursion!
  FOR rowOffset = -1 TO 1    ' Check edges (of blanks) for eligible tiles to expose
    FOR colOffset = -1 TO 1
      IF rowOffset = 0 AND colOffset = 0 THEN ITERATE FOR ' Pointing to Self (endless loop)
      IF rowPtr + rowOffset > 0 AND _                     ' Don't work outside the game board.
         colPtr + colOffset > 0 AND _
         rowPtr + rowOffset <= gGameInfo.gameRows AND _
         colPtr + colOffset <= gGameInfo.gameCols THEN
        CALL ExposeToEdge(rowPtr+rowOffset, colPtr+colOffset, allowcheat) ' Recursively check the edges
      END IF
    NEXT colOffset
  NEXT rowOffset

END SUB
'
'------------------------------------
'====================================
SUB MakeBitMap(hWnd AS DWORD, xWidth AS LONG, yHeight AS LONG)
  LOCAL hDC       AS LONG, _
        hPen      AS LONG, _
        hOldPen   AS LONG, _
        hBrush    AS LONG, _
        hOldBrush AS LONG, _
        nLeft     AS LONG, _
        nRight    AS LONG, _
        nTop      AS LONG, _
        nBottom   AS LONG, _
        fudge1    AS LONG, _
        fudge2    AS LONG, _
        penWidth  AS LONG

  DIM pt(0:2) AS LOCAL POINTAPI

  hDC = GetDC(hWnd)
  IF ghMemDCpic <> 0 THEN
    CALL DeleteObject (SelectObject (ghMemDCpic, ghOldBitmapPic ))
    CALL DeleteObject (ghMemDCpic)
  END IF
  ghMemDCpic     = CreateCompatibleDC(%NULL)
  ghBitmapPic    = CreateCompatibleBitmap(hDC, xWidth*4, yHeight)
  ghOldBitmapPic = SelectObject(ghMemDCpic, ghBitmapPic)
  CALL ReleaseDC(hWnd, hDC)

' Set up variable measures
  fudge1   = yHeight * .16   ' Fudge Factor 1
  fudge2   = yHeight * .05   ' Fudge Factor 2
  penWidth = yHeight * .08
  nTop     = fudge1
  nLeft    = nTop
  nBottom  = yHeight - fudge1
  nRight   = nBottom

' Construct image1 - Bomb on White Background
' Draw Background White Brush and White Pen
  hPen = CreatePen(%PS_SOLID,1, RGB(255,255,255) )
  hBrush = CreateSolidBrush(RGB(255,255,255) )
  hOldPen = SelectObject(ghMemDCpic, hPen)
  hOldBrush = SelectObject(ghMemDCpic, hBrush)
  CALL Rectangle(ghMemDCpic, 0, 0, xWidth+1, yHeight)
 ' Black Pen for bomb
  hPen = CreatePen(%PS_SOLID,1, %BLACK_PEN )
  hBrush = CreateSolidBrush(RGB(0,0,0) )
  CALL SelectObject(ghMemDCpic, hPen)
  CALL SelectObject(ghMemDCpic, hBrush)
  CALL Ellipse(ghMemDCpic, nLeft, nTop, nRight, nBottom)
  hPen = CreatePen(%PS_SOLID,penWidth, %BLACK_PEN )
  CALL SelectObject(ghMemDCpic, hPen)
  CALL MoveTo (ghMemDCpic, xWidth\2, fudge2)  'Up/Down
  CALL LineTo (ghMemDCpic, xWidth\2, yHeight-fudge2)
  CALL MoveTo (ghMemDCpic, fudge2, yHeight\2)
  CALL LineTo (ghMemDCpic, xWidth-fudge2, yHeight\2) 'Left/Right
  CALL MoveTo (ghMemDCpic, nLeft-1, nTop)
  CALL LineTo (ghMemDCpic, nRight, nBottom)  ' Upper Left/Lower Right
  CALL MoveTo (ghMemDCpic, nLeft-1, nBottom)
  CALL LineTo (ghMemDCpic, nRight, nTop)' LowerLeft/UpperRight
  hPen = CreatePen(%PS_SOLID,1, RGB(255,255,255))
  hBrush = CreateSolidBrush(RGB(255,255,255) )
  CALL SelectObject(ghMemDCpic, hPen)
  CALL SelectObject(ghMemDCpic, hBrush)
  CALL Ellipse(ghMemDCpic, nLeft + nRight\8 + fudge2, nTop + nBottom\8 + fudge2, _
               nRight\2 + fudge2, nBottom\2 + fudge2)

' Construct image2 - Bomb on Red Background
' Draw Background Red Brush and Red Pen
  hPen = CreatePen(%PS_SOLID,1, RGB(255,0,0) )
  hBrush = CreateSolidBrush(RGB(255,0,0) )
  CALL SelectObject(ghMemDCpic, hPen)
  CALL SelectObject(ghMemDCpic, hBrush)
  CALL Rectangle(ghMemDCpic,xWidth, 0, xWidth * 2+1, yHeight) 'Left,Top,Right,Bottom
 ' Black Pen for bomb
  hPen = CreatePen(%PS_SOLID,1, %BLACK_PEN )
  hBrush = CreateSolidBrush(RGB(0,0,0) )
  CALL SelectObject(ghMemDCpic, hPen)
  CALL SelectObject(ghMemDCpic, hBrush)
  CALL Ellipse(ghMemDCpic, xWidth + nLeft, nTop, xWidth + nRight, nBottom) 'Left,Top,Right,Bottom
  hPen = CreatePen(%PS_SOLID,penWidth, %BLACK_PEN )
  CALL SelectObject(ghMemDCpic, hPen)
  CALL MoveTo (ghMemDCpic, xWidth + xWidth\2, fudge2)
  CALL LineTo (ghMemDCpic, xWidth + xWidth\2, yHeight-fudge2)  ' Up/down
  CALL MoveTo (ghMemDCpic, xWidth+fudge2, yHeight\2)
  CALL LineTo (ghMemDCpic, xWidth*2-fudge2, yHeight\2)  ' Left/Right
  CALL MoveTo (ghMemDCpic, xWidth + nLeft-1, nTop)
  CALL LineTo (ghMemDCpic, xWidth + nRight, nBottom)  ' Upper Left/Lower Right
  CALL MoveTo (ghMemDCpic, xWidth + nLeft-1, nBottom)
  CALL LineTo (ghMemDCpic, xWidth + nRight, nTop)' LowerLeft/UpperRight
  hPen = CreatePen(%PS_SOLID,1, RGB(255,255,255))
  hBrush = CreateSolidBrush(RGB(255,255,255) )
  CALL SelectObject(ghMemDCpic, hPen)
  CALL SelectObject(ghMemDCpic, hBrush)
  CALL Ellipse(ghMemDCpic, xWidth + nLeft + nRight\8 + fudge2, nTop + nBottom\8 + fudge2, _
                       xWidth + nRight\2 +fudge2, nBottom\2 + fudge2)

' Construct image3 - Flag on gray background
  pt(0).x = xWidth*2 + fudge1 : pt(0).y = yHeight - fudge2
  pt(1).x = xWidth*2 + xWidth\2 : pt(1).y = yHeight - fudge1*1.4
  pt(2).x = xWidth*3 - fudge1 - 1 : pt(2).y = yHeight - fudge2
' Gray Pen on Gray Background
  hPen = CreatePen(%PS_SOLID,1, RGB(192,192,192) )
  hBrush = CreateSolidBrush(RGB(192,192,192) )
  CALL SelectObject(ghMemDCpic, hPen)
  CALL SelectObject(ghMemDCpic, hBrush)
  CALL Rectangle(ghMemDCpic,xWidth*2, 0, xWidth * 4+1, yHeight) 'Left,Top,Right,Bottom
' Black Pen for flagpole
  hPen = CreatePen(%PS_SOLID,penWidth, %BLACK_PEN )
  CALL SelectObject(ghMemDCpic, hPen)
  CALL MoveTo (ghMemDCpic, xWidth*2 + xWidth\2, fudge1)  'Up/Down
  CALL LineTo (ghMemDCpic, xWidth*2 + xWidth\2, yHeight-fudge1)
' Green Brush for Flagpole Base
  hPen = CreatePen(%PS_SOLID, 1, %BLACK_PEN) 'RGB(16,160,40) )
  CALL SelectObject(ghMemDCpic, hPen)
  hBrush = CreateSolidBrush(RGB(16,162,40) )
  CALL SelectObject(ghMemDCpic, hBrush)
  CALL Polygon (ghMemDCpic, pt(0),3)
' Red Brush for Flag
  hPen = CreatePen(%PS_SOLID, 1, RGB(255,0,0) )
  CALL SelectObject(ghMemDCpic, hPen)
  hBrush = CreateSolidBrush(RGB(255,0,0) )
  CALL SelectObject(ghMemDCpic, hBrush)
  CALL Rectangle(ghMemDCpic,xWidth*2 + xWidth\2 + penWidth\2, fudge1, _
                 xWidth*3-fudge1, fudge1 * 2.5) 'Left,Top,Right,Bottom

' Construct image4 - X-ed Flag on gray background
  pt(0).x = xWidth*3 + fudge1 : pt(0).y = yHeight - fudge2
  pt(1).x = xWidth*3 + xWidth\2 : pt(1).y = yHeight - fudge1*1.4
  pt(2).x = xWidth*4 - fudge1 - 1 : pt(2).y = yHeight - fudge2
' Gray Pen on Gray Background
  hPen = CreatePen(%PS_SOLID,1, RGB(192,192,192) )
  hBrush = CreateSolidBrush(RGB(192,192,192) )
  CALL SelectObject(ghMemDCpic, hPen)
  CALL SelectObject(ghMemDCpic, hBrush)
  CALL Rectangle(ghMemDCpic,xWidth*3, 0, xWidth * 5+1, yHeight) 'Left,Top,Right,Bottom
' Black Pen for flagpole
  hPen = CreatePen(%PS_SOLID,penWidth, %BLACK_PEN )
  CALL SelectObject(ghMemDCpic, hPen)
  CALL MoveTo (ghMemDCpic, xWidth*3 + xWidth\2, fudge1)
  CALL LineTo (ghMemDCpic, xWidth*3 + xWidth\2, yHeight-fudge1)
' Green Brush for Flagpole Base
  hPen = CreatePen(%PS_SOLID, 1, %BLACK_PEN) 'RGB(16,160,40) )
  CALL SelectObject(ghMemDCpic, hPen)
  hBrush = CreateSolidBrush(RGB(16,162,40) )
  CALL SelectObject(ghMemDCpic, hBrush)
  CALL Polygon (ghMemDCpic, pt(0),3)
' Red Brush for Flag
  hPen = CreatePen(%PS_SOLID, 1, RGB(255,0,0) )
  CALL SelectObject(ghMemDCpic, hPen)
  hBrush = CreateSolidBrush(RGB(255,0,0) )
  CALL SelectObject(ghMemDCpic, hBrush)
  CALL Rectangle(ghMemDCpic,xWidth*3 + xWidth\2 + penWidth\2, fudge1, _
                 xWidth*4-fudge1, fudge1 * 2.5) 'Left,Top,Right,Bottom
' X out the Flag
  hPen = CreatePen(%PS_SOLID, penWidth, RGB(0,0,0) )
  CALL SelectObject(ghMemDCpic, hPen)
  CALL MoveTo (ghMemDCpic, xWidth*3 + fudge1, fudge1)
  CALL LineTo (ghMemDCpic, xWidth*4 - fudge1, yHeight-fudge1)
  CALL MoveTo (ghMemDCpic, xWidth*4 - fudge1, fudge1)
  CALL LineTo (ghMemDCpic, xWidth*3 + fudge1, yHeight-fudge1)

' Clean Up
  CALL DeleteObject (SelectObject(ghMemDCpic, hOldPen) )
  CALL DeleteObject (SelectObject(ghMemDCpic, hOldBrush) )

END SUB
'
'------------------------------------
'====================================
SUB PaintLED(BYVAL inValue AS LONG, BYVAL xWidth AS LONG, BYVAL yHeight AS LONG)
  LOCAL tString      AS STRING, _
        i            AS LONG
  DIM tPosition(1:3) AS LOCAL LONG
' Wrap values over 999
  DO WHILE inValue > 999
    inValue = inValue - 999
  LOOP
  tString = FORMAT$(inValue,"000")
  tPosition(1) = VAL(LEFT$(tString,1))
  tPosition(2) = VAL(MID$(tString,2,1))
  tPosition(3) = VAL(RIGHT$(tString,1))
  FOR i = 1 TO 3
    CALL BitBlt(ghMemDCDisp, (i-1)*xWidth, 0, xWidth, yHeight, _
                ghMemDCnum, tPosition(i)*xWidth, 0, %SRCCOPY)
  NEXT i
END SUB
'
'------------------------------------
'====================================
SUB MakeLEDBitMap(hWnd AS DWORD, BYVAL xWidth AS LONG, BYVAL yHeight AS LONG)
  LOCAL hDC          AS LONG, _
        hPen         AS LONG, _
        hOldPen      AS LONG, _
        hBrush       AS LONG, _
        hOldBrush    AS LONG, _
        PaintBitInfo AS PaintBitInfoTYPE


  hPen   = 0
  hBrush = 0
  hDC            = GetDC(hWnd)
  ghMemDCnum     = CreateCompatibleDC(%NULL)
  ghBitmapNum    = CreateCompatibleBitmap(hDC, xWidth*11, yHeight)
  ghOldBitmapNum = SelectObject(ghMemDCnum, ghBitmapNum)

' Draw Background: Black Brush and Black Pen
  hPen   = CreatePen(%PS_SOLID,1, RGB(0,0,0) )
  hBrush = CreateSolidBrush(RGB(0,0,0) )
  hOldPen = SelectObject(ghMemDCnum, hPen)
  hOldBrush = SelectObject(ghMemDCnum, hBrush)
  CALL Rectangle(ghMemDCnum, 0, 0, xWidth*11+1, yHeight)
 ' Red Pen for LED
  hPen = CreatePen(%PS_SOLID,1, RGB(255,0,0) )
  hBrush = CreateSolidBrush(RGB(255,0,0) )
  CALL SelectObject(ghMemDCnum, hPen)
  CALL SelectObject(ghMemDCnum, hBrush)
  PaintBitInfo.barThickness = xWidth \ 12
  PaintBitInfo.nWidth  = xWidth
  PaintBitInfo.nHeight = yHeight
  CALL DrawLEDbars(PaintBitInfo, BYVAL 0, BYVAL %Lzero)
  CALL DrawLEDbars(PaintBitInfo, BYVAL 1, BYVAL %Lone)
  CALL DrawLEDbars(PaintBitInfo, BYVAL 2, BYVAL %Ltwo)
  CALL DrawLEDbars(PaintBitInfo, BYVAL 3, BYVAL %Lthree)
  CALL DrawLEDbars(PaintBitInfo, BYVAL 4, BYVAL %Lfour)
  CALL DrawLEDbars(PaintBitInfo, BYVAL 5, BYVAL %Lfive)
  CALL DrawLEDbars(PaintBitInfo, BYVAL 6, BYVAL %Lsix)
  CALL DrawLEDbars(PaintBitInfo, BYVAL 7, BYVAL %Lseven)
  CALL DrawLEDbars(PaintBitInfo, BYVAL 8, BYVAL %Leight)
  CALL DrawLEDbars(PaintBitInfo, BYVAL 9, BYVAL %Lnine)

' Clean Up
  IF hPen   > 0 THEN CALL DeleteObject (SelectObject(ghMemDCnum, hOldPen) )
  IF hBrush > 0 THEN CALL DeleteObject (SelectObject(ghMemDCnum, hOldBrush) )

  CALL ReleaseDC(hWnd, hDC)
END SUB
'
'------------------------------------
'====================================
SUB DrawLEDbars(PBI AS PaintBitInfoTYPE, BYVAL xFigure AS LONG, BYVAL xBitPattern AS BYTE )

  DIM pt(0:7) AS LOCAL POINTAPI
  LOCAL padding            AS LONG, _     ' blank space around the edge of the number
        halfHeight         AS LONG, _     ' rather than recalculating this value, do it once
        centerBarThickness AS LONG, _     ' for balance, the center LED has to be thicker than the others
        cbUpDown           AS LONG, _     ' center Bar, number of pixels above/below center
        offset             AS LONG        ' Offset within the bitmap

  offset = xFigure * PBI.nWidth
  padding = MAX&(PBI.barThickness\4, 1)
  centerBarThickness = 1.5 * PBI.barThickness
' balance center bar by making it an odd number.
  IF INT(centerBarThickness\2)*2 = centerBarThickness THEN INCR centerBarThickness
  cbUpDown   = (centerBarThickness -1)\2
  halfHeight = PBI.nHeight\2
'Upper left bar
  IF BIT(xBitPattern, 1) THEN
    pt(0).x = offset + padding + 1       : pt(0).y = padding + 2
    pt(1).x = pt(0).x                    : pt(1).y = halfHeight - 1
    pt(2).x = pt(1).x + PBI.barThickness : pt(2).y = pt(1).y - PBI.barThickness
    pt(3).x = pt(2).x                    : pt(3).y = pt(0).y + PBI.barThickness
    CALL Polygon (ghMemDCnum, pt(0),4)
  END IF
'Upper right bar
  IF BIT(xBitPattern, 2) THEN
    pt(0).x = offset + PBI.nWidth - (padding + 1) : pt(0).y = padding + 2
    pt(1).x = pt(0).x                             : pt(1).y = halfHeight - 1
    pt(2).x = pt(1).x - PBI.barThickness          : pt(2).y = pt(1).y - PBI.barThickness
    pt(3).x = pt(2).x                             : pt(3).y = pt(0).y + PBI.barThickness
    CALL Polygon (ghMemDCnum, pt(0),4)
  END IF
'Lower left bar
  IF BIT(xBitPattern, 3) THEN
    pt(0).x = offset + padding + 1       : pt(0).y = HalfHeight + 1
    pt(1).x = pt(0).x + PBI.barThickness : pt(1).y = pt(0).y + PBI.barThickness
    pt(2).x = pt(1).x                    : pt(2).y = PBI.nheight - (padding + 2 + PBI.barThickness)
    pt(3).x = pt(0).x                    : pt(3).y = pt(2).y + PBI.barThickness
    CALL Polygon (ghMemDCnum, pt(0),4)
  END IF
'Lower right bar
  IF BIT(xBitPattern, 4) THEN
    pt(0).x = offset + PBI.nWidth - (padding + 1) : pt(0).y = HalfHeight + 1
    pt(1).x = pt(0).x                             : pt(1).y = PBI.nHeight - (padding + 2)
    pt(2).x = pt(1).x - PBI.barThickness          : pt(2).y = PBI.nheight - (padding + 2 + PBI.barThickness)
    pt(3).x = pt(2).x                             : pt(3).y = pt(0).y + PBI.barThickness
    CALL Polygon (ghMemDCnum, pt(0),4)
  END IF
'Top bar
  IF BIT(xBitPattern, 5) THEN
    pt(0).x = offset + padding + 2                : pt(0).y = padding + 1
    pt(1).x = offset + PBI.nwidth - (padding + 2) : pt(1).y = pt(0).y
    pt(2).x = pt(1).x - PBI.barThickness          : pt(2).y = pt(0).y + PBI.barThickness
    pt(3).x = pt(0).x + PBI.barThickness          : pt(3).y = pt(2).y
    CALL Polygon (ghMemDCnum, pt(0),4)
  END IF
'Center bar
  IF BIT(xBitPattern, 6) THEN
    pt(0).x = offset + padding + 2                : pt(0).y = HalfHeight
    pt(3).x = offset + PBI.nWidth - (padding + 2) : pt(3).y = pt(0).y
    pt(1).x = pt(0).x + cbUpDown                  : pt(1).y = pt(0).y - cbUpDown
    pt(2).x = pt(3).x - cbUpDown                  : pt(2).y = pt(3).y - cbUpDown
    pt(4).x = pt(3).x - cbUpDown                  : pt(4).y = pt(3).y + cbUpDown
    pt(5).x = pt(0).x + cbUpDown                  : pt(5).y = pt(0).y + cbUpDown
    CALL Polygon (ghMemDCnum, pt(0),6)
  END IF
'Bottom bar
  IF BIT(xBitPattern, 7) THEN
    pt(0).x = offset + padding + 2                : pt(0).y = PBI.nHeight - (padding + 1)
    pt(1).x = offset + PBI.nwidth - (padding + 2) : pt(1).y = pt(0).y
    pt(2).x = pt(1).x - PBI.barThickness          : pt(2).y = pt(0).y - PBI.barThickness
    pt(3).x = pt(0).x + PBI.barThickness          : pt(3).y = pt(2).y
    CALL Polygon (ghMemDCnum, pt(0),4)
  END IF
END SUB
'
'------------------------------------
'====================================
SUB ReadINI(szINIPath AS ASCIIZ)

  LOCAL bufferSize    AS LONG, _
        fhandle       AS LONG, _
        szSectionName AS ASCIIZ * 16,_
        szKeyName     AS ASCIIZ * 12,_
        szKeyResult   AS ASCIIZ * 8, _
        szKeyDefault  AS ASCIIZ * 8

  fhandle    = 0
  bufferSize = 10

  IF DIR$(szINIPath) = "" THEN
' Defaults
    gGameInfo.gameCols   = 10
    gGameInfo.gameRows   = 10
    gGameInfo.noBombs    = 10
    gGameInfo.buttonSize = 1000
    gGameInfo.noTiles    = gGameInfo.gameCols * gGameInfo.gameRows
    CALL WriteINI(szINIPath)
    EXIT SUB
  END IF

  szSectionName = "GamePreferences"
' ~~~~~~~~~~~~~
  szKeyName     = "Columns"
  szKeyDefault = "10"
  CALL GetPrivateProfileString(szSectionName, szKeyName, szKeyDefault ,szKeyResult, bufferSize, szINIPath)
  gGameInfo.gameCols = VAL(szKeyResult)

  szKeyName     = "Rows"
  CALL GetPrivateProfileString(szSectionName, szKeyName, szKeyDefault ,szKeyResult, bufferSize, szINIPath)
  gGameInfo.gameRows = VAL(szKeyResult)

  szKeyName     = "Bombs"
  CALL GetPrivateProfileString(szSectionName, szKeyName, szKeyDefault ,szKeyResult, bufferSize, szINIPath)
  gGameInfo.noBombs = VAL(szKeyResult)

  szKeyName     = "ButtonSize"
  szKeyDefault = "1000"
  CALL GetPrivateProfileString(szSectionName, szKeyName, szKeyDefault ,szKeyResult, bufferSize, szINIPath)
  gGameInfo.buttonSize = VAL(szKeyResult)

  CLOSE fhandle
  gGameInfo.noTiles = gGameInfo.gameCols * gGameInfo.gameRows

END SUB
'
'------------------------------------
'====================================
SUB WriteINI(szINIPath AS ASCIIZ)
  LOCAL fhandle       AS LONG, _
        szSectionName AS ASCIIZ * 16,_
        szKeyName     AS ASCIIZ * 12,_
        szString      AS ASCIIZ * 8

' Check for a existence of file/write-protected disk/Permission denied first! Exit routine if found.
  ON ERROR GOTO WriteIniError
  fhandle = FREEFILE
  IF DIR$(szINIPath) = "" THEN  ' Avoid a file not found error on INPUT.
    OPEN szINIPath FOR OUTPUT AS fhandle
  ELSE
    OPEN szINIPath FOR INPUT AS fhandle
  END IF
  CLOSE fhandle

  szSectionName = "GamePreferences"
' ~~~~~~~~~~~~~
  szKeyName     = "Columns"
  szString = FORMAT$(gGameInfo.gameCols)
  IF WritePrivateProfileString(szSectionName, szKeyName,szString, szINIPath) = 0 THEN
    EXIT SUB
  END IF
  szKeyName     = "Rows"
  szString = FORMAT$(gGameInfo.gameRows)
  IF WritePrivateProfileString(szSectionName, szKeyName,szString, szINIPath) = 0 THEN
    EXIT SUB
  END IF
  szKeyName     = "Bombs"
  szString = FORMAT$(gGameInfo.noBombs)
  IF WritePrivateProfileString(szSectionName, szKeyName,szString, szINIPath) = 0 THEN
    EXIT SUB
  END IF
  szKeyName     = "ButtonSize"
  szString = FORMAT$(gGameInfo.buttonSize)
  IF WritePrivateProfileString(szSectionName, szKeyName,szString, szINIPath) = 0 THEN
    EXIT SUB
  END IF
  GOTO WriteIniExit

WriteIniError:
'~~~~~~~~~~~~~
  fhandle = ERRCLEAR

WriteIniExit:
'~~~~~~~~~~~~
  ON ERROR GOTO 0

END SUB

'
'---- End Of File -------------------
'====================================
